#ifndef __FIFO_H__
#define __FIFO_H__


#include <stddef.h>
#include <malloc.h>


typedef struct strct_fifo {
	struct strct_fifo *next;
	void *obj;
} fifo_t;


static inline void __attribute__((always_inline)) fifo_print(fifo_t *f) {
	int found = 0;
	while ( f->next != NULL) {
		printf("%p ", f->next);
		f = f->next;
		found = 1;
	}

	if ( found == 1 ) {
		printf("\n");
	}
	
}


static inline fifo_t* __attribute__((always_inline)) fifo_create() {
	fifo_t *f = malloc( sizeof(fifo_t) );
	f->next = NULL;
	f->obj = NULL;

	return f;
}


static inline __attribute__((always_inline)) void fifo_push(fifo_t *f, void* obj) {
	fifo_t *e = malloc( sizeof(fifo_t) );
	e->obj = obj;
	e->next = NULL;

	fifo_t *back = f->obj;
	if ( back != NULL ) {
		back->next = e;
	} else {
		f->next = e;
	}
	f->obj = e;
}


static inline void* __attribute__((always_inline)) fifo_pop(fifo_t *f) {
	fifo_t *front = f->next;
	void * res = NULL;
	if ( front != NULL ) {
		f->next = front->next;

		res = front->obj;
		free( front );

		if ( f->next == NULL ) {
			f->obj = NULL;
		}
	}

	return res;
}


static inline int __attribute__((always_inline)) fifo_isempty(fifo_t *f) {
	return f->next==NULL;
}


static inline void __attribute__((always_inline)) fifo_free(fifo_t *f) {
	while ( f->next != NULL ) {
		fifo_t *victim = f->next;
		f->next = victim->next;

		free(victim);
	}

	free(f);
}


#endif // __FIFO_H__
