#ifndef __STACK_H__
#define __STACK_H__


#include <stddef.h>
#include <malloc.h>

#define STACK_DEFSIZE 1

typedef struct strct_sbody { 
	void* elem[STACK_DEFSIZE];
	struct strct_sbody *ext;
	struct strct_sbody *prev;
	unsigned int size;
	int top;
} sbody_t;


typedef struct strct_stack {
	sbody_t *back;
	size_t elemc;
} stack_t;


#if 1
static inline void __attribute__((always_inline)) stack_print(stack_t *s) {
	sbody_t * b = s->back;
	size_t i = 0;
	size_t n = s->elemc;
	printf("size of stack %p = %zi back %p\n", s, n, b);
	while ( i < n) {
		int j;
		for ( j=b->top-1; j>-1;j--) {
			printf("%zi %p %p ", i+j, b, b->elem[j]);
		//	fflush(0);
		}
		i += b->size;
		b = b->prev;
	//	printf("nextb %p\n", b);
	}

	if ( n != 0 ) {
		printf("\n");
	}
	
}



static inline stack_t* __attribute__((always_inline)) stack_create() {
	stack_t *s = (stack_t*) malloc(sizeof(stack_t));
	sbody_t *body = (sbody_t*) malloc(sizeof(sbody_t));

	body->ext = NULL;
	body->size = STACK_DEFSIZE;
	body->prev = NULL;
	body->top = -1;

	s->back = body;
	s->elemc = 0;

	return s;
}



static inline void __attribute__((always_inline)) stack_push(stack_t *s, void* obj) {
	sbody_t *b = s->back;	
	s->elemc++;

	//printf("stack push %li to %p\n", obj, s);

	unsigned int next = b->top + 1;
	if ( next == b->size ) {
		sbody_t *c = (sbody_t*) malloc(sizeof(sbody_t));
		//printf("allocating %p\n", c);
		c->ext = NULL;
		c->size = STACK_DEFSIZE;
		c->prev = b;
		c->top = -1;

		s->back = b->ext = c;
		b = c;
	} 

	b->elem[++b->top] = obj;
	//printf("added %p to %p at %p\n", obj, s, b);
}


static inline void __attribute__((always_inline)) stack_init(stack_t *s, void *obj[], unsigned int c) {
	unsigned int i;
	for ( i=0; i<c ; i++ ) {
		stack_push(s,obj[i]);
	}
}


static inline __attribute__((always_inline)) void* stack_pop(stack_t *s) {
	sbody_t *b = s->back;	
	int top = b->top;
	
	if ( top == -1 ) {
		return NULL;
	}

	void *elem = b->elem[top];

	if ( top == 0 ) {
		sbody_t *c = b->prev;
		if ( c != NULL ) {
			c->ext = NULL;
			s->back = c;
			free(b);

			//printf("removing %p\n", b);
		}
	} 

	--s->elemc;

	return elem;
}



static inline int __attribute__((always_inline)) stack_isempty(stack_t *s) {
	return s->elemc==0;
}


static inline void __attribute__((always_inline)) stack_free(stack_t *s) {
//	printf("\nstack_free %p %zi\n", s, s->size);
	//stack_print(s);
	sbody_t *b = s->back;	
//	printf("body %p\n", b);
	while ( b != NULL ) {
		sbody_t *prevb = b->prev;
//		printf("prevb %p\n", prevb);
//		fflush(0);
		free(b);
//		printf("removed %p\n", b);
//		fflush(0);
		b = prevb;
	}
	free(s);
}


#endif


#endif // __STACK_H__
