#ifndef __VECTOR_H__
#define __VECTOR_H__


#include <stddef.h>
#include <malloc.h>


#define VECT_DEFSIZE	128

typedef union struc_velem {
	void *p;
	int i;
	double d;
} vel_t;

typedef struct struct_vect {
	vel_t* elem;
	size_t elemc;
	size_t size;
} vector_t;


static inline vector_t* __attribute__((always_inline)) vector_printi(vector_t *v) {
	vel_t *elem = v->elem;
	size_t elemc = v->elemc;

	int i;
	for ( i=0; i<elemc; i++ ) {
		printf("%i ", elem[i].i);
	}	

	if ( elemc ) {
		printf("\n");
	}
}


static inline vector_t* __attribute__((always_inline)) vector_printd(vector_t *v) {
	vel_t *elem = v->elem;
	size_t elemc = v->elemc;

	int i;
	for ( i=0; i<elemc; i++ ) {
		printf("%.4f ", elem[i].d);
	}	

	if ( elemc ) {
		printf("\n");
	}
}


static inline int __attribute__((always_inline)) vector_contains(vector_t *v, vel_t obj) {
	size_t elemc = v->elemc;
	vel_t *elem = v->elem;
	
	int i;
	for ( i=0; i<elemc; i++ ) {
		if ( elem[i].i == obj.i ) {
			return 1;
		}
	}
	
	return 0;
}

static inline vector_t* __attribute__((always_inline)) vector_create() {
	vector_t *v = (vector_t*) malloc(sizeof(vector_t));
	vel_t *elem = (vel_t*) malloc(sizeof(vel_t) * VECT_DEFSIZE);

	v->elem = elem;
	v->elemc = 0;
	v->size = VECT_DEFSIZE;

	return v;
}




static inline void __attribute__((always_inline)) vector_free(vector_t *v) {
	if ( v->elem != NULL ) {
		free(v->elem);
	}
	free(v);
}


static inline vel_t __attribute__((always_inline)) vector_get(vector_t *v, size_t i) {
	return v->elem[i];
}


static inline void __attribute__((always_inline)) vector_insertat(vector_t *v, vel_t obj, size_t i) {
	if ( i >= v->size ) {
		size_t newsize = v->size + VECT_DEFSIZE;
		v->elem = realloc(v->elem, sizeof(vel_t) * newsize);
		v->size = newsize;
	}

	size_t s = i+1;
	if ( s > v->elemc ) {
		v->elemc = s;
	}

	v->elem[i] = obj;
}

static inline void __attribute__((always_inline)) vector_insert(vector_t *v, vel_t obj) {
	vector_insertat(v, obj, v->elemc);
}


static inline int __attribute__((always_inline)) vector_insert_t(vector_t *v, vel_t obj) {
	if ( ! vector_contains(v, obj) ) {
		vector_insertat(v, obj, v->elemc);
		return 1;
	}

	return 0;
}


static inline vector_t* __attribute__((always_inline)) vector_appendi(vector_t *v, int *objs, size_t size) {
	int i;
	for ( i=0; i< size; i++ ) {
		vel_t e;
		e.i = objs[i];
		vector_insert(v, e);
	}

	return v;
}


static inline vector_t* __attribute__((always_inline)) vector_append(vector_t *v, vel_t *objs, size_t size) {
	int i;
	for ( i=0; i< size; i++ ) {
		vector_insert(v, objs[i]);
	}

	return v;
}


static inline vector_t* __attribute__((always_inline)) vector_clear(vector_t *v) {
	v->elemc = 0;
}


static inline void __attribute__((always_inline)) vector_rm(vector_t *v, vel_t obj) {
	vel_t* elem = v->elem;
	size_t elemc = v->elemc;

	int i=-1;
	int found = 0;
	while ( !found ) {
		++i;
		found = elem[i].i==obj.i;
	}

	int j;
	for ( j=i; j<elemc; j++ ) {
		elem[j] = elem[j+1];
	}

	if ( i == elemc - 1) {
		--v->elemc;
	}
}


static inline int* __attribute__((always_inline)) vector2int(vector_t *v) {
	vel_t *elem = v->elem;
	size_t elemc = v->elemc;
	int *intelem = (int*) malloc( v->elemc * sizeof(int) );

	int i;
	for ( i=0; i < elemc ; i++ ) {
		intelem[i] = elem[i].i;
	}
	
	vector_free(v);

	return intelem;
}


static inline double* __attribute__((always_inline)) vector2double(vector_t *v) {
	vel_t *elem = v->elem;
	size_t elemc = v->elemc;
	double *delem = (double*) malloc( v->elemc * sizeof(double) );

	int i;
	for ( i=0; i < elemc ; i++ ) {
		delem[i] = elem[i].d;
	}
	
	vector_free(v);

	return delem;
}


#endif // __VECTOR_H__
