#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>


#include "matmul_config.h"
#include "matmul_setup.h"
#include "matmul_check.h"
#if USE_STRASSEN
#include "strassen_main.h"
#elif USE_WINOGRAD
#include "winograd_main.h"
#elif USE_MM
#include "mm.h"
#endif
#include "fptype.h"


/* user dimensions */
int m;
int k;
int n;
/* rounded dimensions */
int rm;
int rk;
int rn;
/* block dimensions */
int b;
int c;
int d;
/* level of recursion */
unsigned int l;
int reps;
int check;


fp_t *A;
fp_t *B;
fp_t *C;

fp_t *S;
unsigned long sscratchs;


int main(int argc, char* argv[]) {
	if ( matmul_config(argc, argv) ) {
		return 1;
	}

	if ( matmul_setup( check, m, rm, n, rn, k, rk, b, d, c) ) {
		return 2;
	}

	unsigned long elapsed=0;
	int r;
	for ( r = 0; r < reps; ++r ) {
		struct timeval start;
		gettimeofday( &start, NULL );

#if USE_STRASSEN
		strassen_main(b, c, d, rm, rk, rn, A, B, C, S);
#elif USE_WINOGRAD
		winograd_main(b, c, d, rm, rk, rn, A, B, C, S);
#elif USE_MM
		mm(b, d, c, rm, rn, rk, A, B, C);
#endif

		#pragma omp taskwait

		struct timeval stop;
		gettimeofday(&stop,NULL);

		unsigned long interm = stop.tv_usec - start.tv_usec;
		interm += (stop.tv_sec - start.tv_sec ) * 1000000;
		elapsed += interm;
  	}

	//fprint_ddense2mm("C.mm", "C", mm, nn, C, mm);

	double elapsedfl = (double) elapsed;

  	FILE *statf = fopen("ompss-stats-0001", "w");
	fprintf(statf,"time : %.2f ms\n", elapsedfl / ((double) reps * 1e3));
	fclose(statf);

	matmul_check(check, b, d, c, rm, rn, rk, A, B, C);

	matmul_shutdown(check, m, rm, n, rn, k, rk, b, d, c);

	return 0;
}
