#ifndef __BBLAS_GEMM_H__
#define __BBLAS_GEMM_H__


#include <stdio.h>
#include <stdlib.h>

#include "fptype.h"
#include "tasks_gemm.h"


#ifdef SINGLE_PRECISION
#define BBLAS_GEMM 			bblas_sgemm
#define BBLAS_HYPER_GEMM 	bblas_hyper_sgemm
#else
#define BBLAS_GEMM 			bblas_dgemm
#define BBLAS_HYPER_GEMM 	bblas_hyper_dgemm
#endif


static inline void __attribute__((always_inline)) bblas_sgemm(int bm, int bn, int bk, int m, int n, int k, float alpha, float *A, int ldimA, float *B, int ldimB, 
	float beta, float *C, int ldimC) {
	int i;
	for ( i=0; i<m; i+= bm ) {
		int j;
		for ( j=0; j<n; j+=bn ) {
			task_sgemm(bm, bn, bk, alpha, &A[0*ldimA+i], ldimA, &B[j*ldimB+0], ldimB, beta, &C[j*ldimC+i], ldimC);
			int l;
			for ( l=bk; l<k; l+=bk) {
				task_sgemm(bm, bn, bk, alpha, &A[l*ldimA+i], ldimA, &B[j*ldimB+l], ldimB, FP_ONE, &C[j*ldimC+i], ldimC);
			}
		}
	}
}

static inline void __attribute__((always_inline)) bblas_hyper_sgemm(int bm, int bn, int bk, int m, int n, int k, float alpha, float *A, int ldimA, float *B, int ldimB, 
	float beta, float *C, int ldimC) {
	int i;
	for ( i=0; i<m; i+= bm ) {
		int j;
		for ( j=0; j<n; j+=bn ) {
			task_sgemm(bm, bn, bk, alpha, &A[0*ldimA+i], bm, &B[j*ldimB+0], bk, beta, &C[j*ldimC+i], bm);
			int l;
			for ( l=bk; l<k; l+=bk) {
				task_sgemm(bm, bn, bk, alpha, &A[l*ldimA+i], bm, &B[j*ldimB+l], bk, FP_ONE, &C[j*ldimC+i], bm);
			}
		}
	}
}

static inline void __attribute__((always_inline)) bblas_dgemm(int bm, int bn, int bk, int m, int n, int k, double alpha, double *A, int ldimA, double *B, int ldimB, 
	double beta, double *C, int ldimC) {
	int i;
	for ( i=0; i<m; i+= bm ) {
		int j;
		for ( j=0; j<n; j+=bn ) {
			task_dgemm(bm, bn, bk, alpha, &A[0*ldimA+i], ldimA, &B[j*ldimB+0], ldimB, beta, &C[j*ldimC+i], ldimC);

			int l;
			for ( l=bk; l<k; l+=bk) {
				task_dgemm(bm, bn, bk, alpha, &A[l*ldimA+i], ldimA, &B[j*ldimB+l], ldimB, FP_ONE, &C[j*ldimC+i], ldimC);
			}
		}
	}
}

static inline void __attribute__((always_inline)) bblas_hyper_dgemm(int bm, int bn, int bk, int m, int n, int k, double alpha, double *A, int ldimA, double *B, int ldimB, 
	double beta, double *C, int ldimC) {
	int i;
	for ( i=0; i<m; i+= bm ) {
		int j;
		for ( j=0; j<n; j+=bn ) {
			task_dgemm(bm, bn, bk, alpha, &A[0*ldimA+i*bk], bm, &B[j*ldimB+0], bk, beta, &C[j*ldimC+i*bn], bm);
			int l;
			for ( l=bk; l<k; l+=bk) {
				task_dgemm(bm, bn, bk, alpha, &A[l*ldimA+i*bk], bm, &B[j*ldimB+l*bn], bk, FP_ONE, &C[j*ldimC+i*bn], bm);
			}
		}
	}
}


#endif // __BBLAS_GEMM_H__
