#include "genmat.h"

#include "fptype.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>


#ifdef DOUBLE_PRECISION
#define __genmat 				dgenmat
#define __genmat_dd 			dgenmat_dd
#define __genmat_covariant 		dgenmat_covariant
#else
#define __genmat 				sgenmat
#define __genmat_dd 			sgenmat_dd
#define __genmat_covariant 		sgenmat_covariant
#endif


void *__genmat(int m, int n) {
	fp_t *A = malloc(m * n * sizeof(fp_t));
	int mn = m * n;

	int j;
	for (j=0; j<mn ; ++j) {
		A[j] = FP_RAND();
	}

	return A;
}


void *__genmat_dd(int n) {
	FP_SEED( ( long int ) time(NULL) );
	
	fp_t *A = malloc(n * n * sizeof(fp_t));
	int j;
	for (j=0; j<n ; ++j) {
		fp_t *ap = &A[j*n+0];

		int i;
		for (i=0; i<j ; ++i) {
			*ap++ = A[i*n+j];
		}    

		fp_t f = FP_RAND() + n;
		*ap++ = f;
		if ( f == 0.0 ) {
			printf("generated 0\n");
		}

		for (i=j+1; i<n ; ++i) {
			fp_t f = FP_RAND();
			*ap++ = f;
			if ( f == 0.0 ) {
				printf("generated 0\n");
			}
		}    
	}

	return A;
}


void *__genmat_covariant(int n, fp_t p, fp_t k) {
	if ( p <= 0.0 ) {
		fprintf(stderr, "err: a covariant matrix requires p > 0\n");
		return NULL;
	}
	
	fp_t mp = -p;
	fp_t kinv = 1 / k;

	fp_t *A = malloc(n * n * sizeof(fp_t));

	int j;
	for (j=0; j<n ; ++j) {
		fp_t *ap = &A[j*n+0];

		int i;
		for (i=0; i<j ; ++i) {
			fp_t v = FP_POW(j-i, mp);
	//		printf("generated (%i,%i) %e\n", i, j, v);
			*ap++ = v;
		}    

		fp_t v = 1 + FP_POW(j+1, kinv);
	//	printf("generated (%i,%i) %e\n", i, j, v);
		*ap++ = v;

		for (i=j+1; i<n ; ++i) {
			fp_t v = FP_POW(i-j, mp);
	//		printf("generated (%i,%i) %e\n", i, j, v);
			*ap++ = v;
		}    
	}

	return A;
}
