#include "morton.h"

#include <string.h>

#include "fifo.h"
#include "quadtree.h"
#include "stack.h"


quadtree_t* morton_qtree(int m, int n, int b, int c) {
	quadtree_t *root = qtree_create_elem(0, m, n);
	root->x = 0;
	root->y = 0;

	stack_t *s = stack_create();
	stack_push( s, root );
	int mortoni = 0;
	int b2 = b * c;

	while ( !stack_isempty( s ) ) {
		quadtree_t *t = stack_pop( s );
		int level = t->level + 1;
		int m = t->m >> 1;
		int n = t->n >> 1;
		int x = t->x;
		int y = t->y;
	
		t->idx = mortoni;

		if ( m < b ) {
			mortoni += b2;
			t->leaf = 1;
		} else {
			quadtree_t *se = t->se = qtree_create_elem(level, m, n);  
			se->x = x+m ; se->y = y+n;
			stack_push( s, se );

			quadtree_t *sw = t->sw = qtree_create_elem(level, m, n);  
			sw->x = x+m ; sw->y = y;
			stack_push( s, sw );

			quadtree_t *ne = t->ne = qtree_create_elem(level, m, n);
			ne->x = x ; ne->y = y+n;
			stack_push( s, ne );

			quadtree_t * nw = t->nw = qtree_create_elem(level, m, n);
			nw->x =x ; nw->y = y;
			stack_push( s, nw );
		}	
	}

	stack_free( s );

	return root;
}


void morton_print_qtree( const char fname[], quadtree_t *t ) { 
	FILE *f = fopen(fname, "w");
	fifo_t *fi = fifo_create();
	fifo_push(fi, t);

	fprintf( f, "digraph {\n");

	t->id = 0;
	int id = 0;

	while ( !fifo_isempty( fi ) ) {
		quadtree_t *t = fifo_pop( fi );
		int tid = t->id;
		//printf("%i: (%i,%i) dim %i %i idx %i\n", tid, t->x, t->y, t->m, t->n, t->idx);
		
		if ( t->leaf == 0) {
			quadtree_t *nw = (quadtree_t*) t->nw;
			fifo_push( fi, nw );
			nw->id = ++id;
			fprintf(f, "%i -> %i;\n", tid, id);

			quadtree_t *ne = (quadtree_t*) t->ne;
			fifo_push( fi, ne );
			ne->id = ++id;
			fprintf(f, "%i -> %i;\n", tid, id);

			quadtree_t *sw = (quadtree_t*) t->sw;
			fifo_push( fi, sw );
			sw->id = ++id;
			fprintf(f, "%i -> %i;\n", tid, id);

			quadtree_t *se = (quadtree_t*) t->se;
			fifo_push( fi, se );
			se->id = ++id;
			fprintf(f, "%i -> %i;\n", tid, id);
		}	
	}

	fifo_free( fi );
	
	fprintf( f, "}\n");
	fclose(f);
}


double* morton2colmajor( int k, int l, int b, int c, double *Am, quadtree_t *t ) {
	double *A = calloc( k * l, sizeof(double) );

	stack_t *s = stack_create();
	stack_push(s, t);

	while ( !stack_isempty( s ) ) {
		quadtree_t *t = stack_pop( s );
		
		if ( t->leaf == 0) {
			stack_push( s, t->se );
			stack_push( s, t->sw );
			stack_push( s, t->ne );
			stack_push( s, t->nw );
		} else {
			int m = t->m;
			int n = t->n;
			int x = t->x;
			int y = t->y;

			double *mblock = &Am[t->idx];	
			//printf("%f %f %f %f\n", mblock[0], mblock[1], mblock[2], mblock[3]);
			double *block = A + y * k + x;
			int j;
			for ( j = 0; j < c; ++j, block+=k, mblock+=b ) {
				memcpy(block, mblock, b * sizeof(double));
			}
		}
	}

	stack_free( s );

	return A;
}
