#include "matmul_check.h"

#include "fptype.h"
#include "fpblas.h"
#include "blocked.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <math.h>


#define max(a,b) (a) > (b)? (a) : (b)
#define min(a,b) (a) > (b)? (b) : (a)


int matmul_check(int check, int b, int d, int c, int m, int n, int k, void *A, void *B, void *C) {
	if ( check ) {
#ifdef HYPER
		fp_t *Acm = BCM2CM(b, c, m, k, A);
		fp_t *Bcm = BCM2CM(c, d, k, n, B);
		fp_t *Ccm = BCM2CM(b, d, m, n, C);
#else
		fp_t *Acm = A;
		fp_t *Bcm = B;
		fp_t *Ccm = C;
#endif
		fp_t *Cx = malloc( sizeof(fp_t) * m * n );

		struct timeval start;
		gettimeofday( &start, NULL );
		BLAS_gemm(MAT_NOTRANSP, MAT_NOTRANSP, m, n, k, FP_ONE, Acm, m, Bcm, k, FP_NOUGHT, Cx, m);
		struct timeval stop;
		gettimeofday(&stop,NULL);
		unsigned long elapsed = stop.tv_usec - start.tv_usec;
		elapsed += (stop.tv_sec - start.tv_sec ) * 1000000;

		fp_t *sumvec = calloc(n, sizeof(fp_t));
		fp_t *sumvecx = calloc(n, sizeof(fp_t));

		int j;
		for ( j=0; j<n; ++j ) {
			int i;
			for ( i=0; i<m; ++i ) {
				fp_t tmp = FP_ABS(Cx[j*m+i]);
				sumvecx[j] += tmp;

				tmp -= FP_ABS(Ccm[j*m+i]);
				sumvec[j] += FP_ABS(tmp);
    			}
		}

		fp_t norm=0.0, normx=0.0;
		for ( j=0; j<n; ++j ) {
			norm = max(norm, sumvec[j]);
			normx = max(normx, sumvecx[j]);
		}

		norm /= normx;

		printf("check: diff norm %e (%.2f ms)\n", norm, elapsed / (double) 1e3 );

#ifdef HYPER
		free( Acm );
		free( Bcm );
		free( Ccm );
#endif
		free( sumvec );
		free( sumvecx );
		free( Cx );
	}

	return 0;
}
