#include "matmul_check.h"


#include <stdlib.h>
#include <time.h>
#include <math.h>

#include "fptype.h"
#include "fpblas.h"
#include "genmat.h"
#include "quadtree.h"
#include "stack.h"
#include "morton.h"

#define max(a,b) (a) > (b)? (a) : (b)
#define min(a,b) (a) > (b)? (b) : (a)


extern unsigned long sscratchs;


int matmul_check(int check, int b, int d, int c, int m, int n, int k, void *A, void *B, void *C) {
	if ( check ) {
		extern unsigned long scratchs;
		if ( sscratchs != scratchs ) {
			fprintf(stderr, "error: incorrect scratch size (%i %i)\n", sscratchs, scratchs);
		}

		quadtree_t *qta = morton_qtree( m, k, b, c);
		double *Acm = morton2colmajor( m, k, b, c, A, qta );	
		quadtree_t *qtb = morton_qtree( k, n, c, d);
		double *Bcm = morton2colmajor( k, n, c, d, B, qtb );	
		quadtree_t *qtc = morton_qtree( m, n, b, d);
		double *Ccm = morton2colmajor( m, n, b, d, C, qtc );	
		//morton_print_qtree( "qta.dot", qta);
		//morton_print_qtree( "qtb.dot", qtb);
		//morton_print_qtree( "qtc.dot", qtc);

		//octave_savemat("a.mat", "A",  m, k, Acm);
		//octave_savemat("a.mat", "A",  m, k, Acm);
		//octave_savemat("b.mat", "B",  k, n, Bcm);
		//octave_savemat("c.mat", "C",  m, n, Ccm);

		double *Cx = malloc( sizeof(double) * m * n );

		struct timeval start;
		gettimeofday( &start, NULL );
		BLAS_gemm(MAT_NOTRANSP, MAT_NOTRANSP, m, n, k, FP_ONE, Acm, m, Bcm, k, FP_NOUGHT, Cx, m);
		struct timeval stop;
		gettimeofday(&stop,NULL);
		unsigned long elapsed = stop.tv_usec - start.tv_usec;
		elapsed += (stop.tv_sec - start.tv_sec ) * 1000000;

		double *sumvec = calloc( n, sizeof(double) );
		double *sumvecx = calloc( n, sizeof(double) );

		//octave_savemat("cex.mat", "Cex",  m, n, Cx);

		int j;
		for (j = 0; j < n; j++) {
			int i;
			for (i = 0; i < m; i++) {
				double tmp = FP_ABS(Cx[j*m+i]);
				sumvecx[j] += tmp;

				tmp -= FP_ABS(Ccm[j*m+i]);
				sumvec[j] += FP_ABS(tmp);
    			}
		}

		double norm=0.0, normx=0.0;
		for ( j=0; j < n; ++j ) {
			norm = max(norm, sumvec[j]);
			normx = max(normx, sumvecx[j]);
		}

		norm /= normx;

		printf("check: diff norm %e (%.2f ms)\n", norm, elapsed / (double) 1e3 );

		free( sumvec );
		free( sumvecx );
		free( Acm );
		free( Bcm );
		free( Ccm );
	}

	return 0;
}
