#ifndef __STRASSEN_P_COMB_H__
#define __STRASSEN_P_COMB_H__


#include "strassen_kernels.h"


// P1 = (A11 + A22) * (B11 + B22)
// P3 = A11 * (B12 - B22) 
// P4 = A22 * (B21 - B11) 
static inline __attribute__((always_inline)) void strassen_P1P34(int b, int c, int d, int m, int k, int n, int bA, int bB, int dimA, int dimB, double *A11, double *A22, double *B11, double *B12, 
	double *B22, double *P1, double *P3, double *P4, double *S, unsigned long *scratchs) {
	unsigned long lscratchs = *scratchs;

	double *F = S + lscratchs;
	lscratchs += dimB;

	int i = 0;
	for ( i = 0; i < dimB; i += bB ) {
		basic_msub(bB, &B12[i], &B22[i], &F[i]); 
	}

	*scratchs = lscratchs;

	strassen_mm(b, c, d, m, k, n, A11, F, P34, S, scratchs);	
	strassen_mm(b, c, d, m, k, n, A11, F, P34, S, scratchs);	
	strassen_mm(b, c, d, m, k, n, A11, F, P34, S, scratchs);	
}


#endif // __STRASSEN_P_COMB_H__
