#include <stdio.h>
#include <stdlib.h>


#include "strassen_kernels.h"
#include "strassen_P.h"
#include "debug.h"


void strassen_mm(int b, int c, int d, int m, int k, int n, double *A, double *B, double *C, double *S, unsigned long *scratchs) {
	int hm = m >> 1;
	int hk = k >> 1;
	int hn = n >> 1;
	int bA = b * c;
	int bB = c * d;
	int bC = b * d;
	int dimA = m * k;
	int dimB = k * n;
	int dimC = m * n;
	int hdimA = dimA >> 2;
	int hdimB = dimB >> 2;
	int hdimC = dimC >> 2;

	if ( m == b ) {
		basic_mm(b, c, d, A, B, C);
	} else {
		double *A11 = A;
		double *A12 = A + hdimA;
		double *A21 = A12 + hdimA;
		double *A22 = A21 + hdimA;

		double *B11 = B;
		double *B12 = B + hdimB;
		double *B21 = B12 + hdimB;
		double *B22 = B21 + hdimB;

		double *C11 = C;
		double *C12 = C + hdimC;
		double *C21 = C12 + hdimC;
		double *C22 = C21 + hdimC;

		unsigned long lscratchs = *scratchs;

		// P1 = (A11 + A22) * (B11 + B22)
		double *P1 = S + lscratchs;
		lscratchs += hdimC;
		double *A11p22 = S + lscratchs;
		lscratchs += hdimA;
		double *B11p22 = S + lscratchs;
		lscratchs += hdimB;
		// P2 = (A21 + A22) * B11
		double *P2 = S + lscratchs;
		lscratchs += hdimC;
		double *A21p22 = S + lscratchs;
		lscratchs += hdimA;
		// P3 = A11 * (B12 - B22) 
		double *P3 = S + lscratchs;
		lscratchs += hdimC;
		double *B12m22 = S + lscratchs;
		lscratchs += hdimB;
		// P4 = A22 * (B21 - B11) stored directly into C21
		double *P4 = C21; 
		double *B21m11 = S + lscratchs;
		lscratchs += hdimB;
		// P5 = (A11 + A12) * B22 stored directly into C12
		double *P5 = C12;
		double *A11p12 = S + lscratchs;
		lscratchs += hdimA;
		// P6 = (A21 - A11) * (B11 + B12) stored directly into C22
		double *P6 = C22;
		double *A21m11 = S + lscratchs;
		lscratchs += hdimA;
		double *B11p12 = S + lscratchs;
		lscratchs += hdimB;
		// P7 = (A12 - A22) * (B21 + B22) stored directly into C11
		double *P7 = C11;
		double *A12m22 = S + lscratchs;
		lscratchs += hdimA;
		double *B21p22 = S + lscratchs;
		lscratchs += hdimB;

		int i;
		for ( i = 0; i < hdimA; i += bA ) {
			// A11 + A22, A21 + A22, A11 + A12, A21 - A11, A12 - A22
			strassen_b(bA, &A11[i], &A12[i], &A21[i], &A22[i], &A11p22[i], &A21p22[i], &A11p12[i], &A21m11[i], &A12m22[i]);
		}

		for ( i = 0; i < hdimB; i += bB ) {
			// B11 + B22, B21 + B22, B11 + B12, B21 - B11, B12 - B22
			strassen_b(bB, &B11[i], &B12[i], &B21[i], &B22[i], &B11p22[i], &B21p22[i], &B11p12[i], &B21m11[i], &B12m22[i]);
		}

		strassen_mm(b, c, d, hm, hk, hn, A11p22, B11p22, P1, S, &lscratchs);	
		strassen_mm(b, c, d, hm, hk, hn, A21p22, B11, P2, S, &lscratchs);
		strassen_mm(b, c, d, hm, hk, hn, A11, B12m22, P3, S, &lscratchs);
		strassen_mm(b, c, d, hm, hk, hn, A22, B21m11, P4, S, &lscratchs);
		strassen_mm(b, c, d, hm, hk, hn, A11p12, B22, P5, S, &lscratchs);
		strassen_mm(b, c, d, hm, hk, hn, A21m11, B11p12, P6, S, &lscratchs);
		strassen_mm(b, c, d, hm, hk, hn, A12m22, B21p22, P7, S, &lscratchs);

		for ( i = 0; i < hdimC; i += bC ) {
			// C11 = P1 + P4 - P5 + P7 
			strassen_C1122(bC, &P1[i], &P4[i], &P5[i], &P7[i]);
			// C22 = P1 + P3 - P2 + P6 
			strassen_C1122(bC, &P1[i], &P3[i], &P2[i], &P6[i]);
			// C21 = P2 + P4, C12 = P3 + P5
			strassen_C1221(bC, &P2[i], &P3[i], &P4[i], &P5[i] );
		}

		*scratchs = lscratchs;
	}
}
