#include <stdio.h>
#include <stdlib.h>


#include "strassen_kernels.h"
#include "strassen_P.h"
#include "debug.h"


void strassen_mm(int b, int c, int d, int m, int k, int n, double *A, double *B, double *C, double *S, unsigned long *scratchs) {
	int hm = m >> 1;
	int hk = k >> 1;
	int hn = n >> 1;
	int bA = b * c;
	int bB = c * d;
	int bC = b * d;
	int dimA = m * k;
	int dimB = k * n;
	int dimC = m * n;
	int hdimA = dimA >> 2;
	int hdimB = dimB >> 2;
	int hdimC = dimC >> 2;

	if ( m == b ) {
		basic_mm(b, c, d, A, B, C);
	} else {
		double *A11 = A;
		double *A12 = A + hdimA;
		double *A21 = A12 + hdimA;
		double *A22 = A21 + hdimA;

		double *B11 = B;
		double *B12 = B + hdimB;
		double *B21 = B12 + hdimB;
		double *B22 = B21 + hdimB;

		double *C11 = C;
		double *C12 = C + hdimC;
		double *C21 = C12 + hdimC;
		double *C22 = C21 + hdimC;

		unsigned long lscratchs = *scratchs;

		// P1 = (A11 + A22) * (B11 + B22)
		double *P1 = S + lscratchs;
		lscratchs += hdimC;
		strassen_P1(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A11, A22, B11, B22, P1, S, &lscratchs);

		// P2 = (A21 + A22) * B11
		double *P2 = S + lscratchs;
		lscratchs += hdimC;
		strassen_P25(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A21, A22, B11, P2, S, &lscratchs);

		// P3 = A11 * (B12 - B22) 
		double *P3 = S + lscratchs;
		lscratchs += hdimC;
		strassen_P34(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A11, B12, B22, P3, S, &lscratchs);

		// P4 = A22 * (B21 - B11) stored directly into C21
		double *P4 = C21; 
		strassen_P34(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A22, B21, B11, P4, S, &lscratchs);

		// P5 = (A11 + A12) * B22 stored directly into C12
		double *P5 = C12;
		strassen_P25(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A11, A12, B22, P5, S, &lscratchs);

		// P6 = (A21 - A11) * (B11 + B12) stored directly into C22
		double *P6 = C22;
		strassen_P67(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A21, A11, B11, B12, P6, S, &lscratchs);

		// P7 = (A12 - A22) * (B21 + B22) stored directly into C11
		double *P7 = C11;
		strassen_P67(b, c, d, hm, hk, hn, bA, bB, hdimA, hdimB, A12, A22, B21, B22, P7, S, &lscratchs);


		int i;
		for ( i = 0; i < hdimC; i += bC ) {
			// C11 = P1 + P4 - P5 + P7 
			strassen_C1122(bC, &P1[i], &P4[i], &P5[i], &P7[i]);
			// C22 = P1 + P3 - P2 + P6 
			strassen_C1122(bC, &P1[i], &P3[i], &P2[i], &P6[i]);
			// C21 = P2 + P4
			// C12 = P3 + P5
			strassen_C1221(bC, &P2[i], &P3[i], &P4[i], &P5[i] );
		}

		*scratchs = lscratchs;
	}
}
