#include <stdio.h>
#include <stdlib.h>


#include "wino_kernels.h"


void wino_mm(int b, int c, int d, int m, int k, int n, double *A, double *B, double *C, double *S, unsigned long *scratchs) {
	int hm = m >> 1;
	int hk = k >> 1;
	int hn = n >> 1;
	int bA = b * c;
	int bB = c * d;
	int bC = b * d;
	int dimA = m * k;
	int dimB = k * n;
	int dimC = m * n;
	int hdimA = dimA >> 2;
	int hdimB = dimB >> 2;
	int hdimC = dimC >> 2;

	if ( m == b ) {
		basic_mm(m, k, n, A, B, C);
	} else {
		double *A11 = A;
		double *A12 = A + hdimA;
		double *A21 = A12 + hdimA;
		double *A22 = A21 + hdimA;

		double *B11 = B;
		double *B12 = B + hdimB;
		double *B21 = B12 + hdimB;
		double *B22 = B21 + hdimB;

		unsigned long lscratchs = *scratchs;

		double *C11 = C;
		double *C12 = C + hdimC;
		double *C21 = C12 + hdimC;
		double *C22 = C21 + hdimC;

		// S1 = A21 + A22     	
		// S2 = S1 - A11  	
		// S3 = A11 - A21 	
		// S4 = A12 - S2  	
		double *S1 = S + lscratchs;
		double *S2 = S1 + hdimA;
		double *S3 = S2 + hdimA;
		double *S4 = S3 + hdimA;
		lscratchs += hdimA << 2;
		int i;
		for ( i=0 ; i < hdimA; i+=bA ) {
			wino_a( bA, &A21[i], &A22[i], &A11[i], &A12[i], &S1[i], &S2[i], &S3[i], &S4[i] );
		}

		// T1 = B12 - B11
		// T2 = B22 - T1
		// T3 = B22 - B12
		// T4 = B21 - T2
		double *T1 = S + lscratchs;
		double *T2 = T1 + hdimB;
		double *T3 = T2 + hdimB;
		double *T4 = T3 + hdimB;
		lscratchs += hdimB << 2;
		for ( i=0 ; i < hdimB; i+=bB ) {
			wino_b( bB, &B12[i], &B11[i], &B22[i], &B21[i], &T1[i], &T2[i], &T3[i], &T4[i] );
		}

		// P1 = A11 * B11
		//double *P1 = S + lscratchs;
		//lscratchs += hdimC;
		double *P1 = C11;
		wino_mm(b, c, d, hm, hk, hn, A11, B11, P1, S, &lscratchs);

		// P2 = A12 * B21
		double *P2 = S + lscratchs;
		lscratchs += hdimC;
		wino_mm(b, c, d, hm, hk, hn, A12, B21, P2, S, &lscratchs);

		// P3 = S1 * T1
		//double *P3 = S + lscratchs;
		//lscratchs += hdimC;
		double *P3 = C22;
		wino_mm(b, c, d, hm, hk, hn, S1, T1, P3, S, &lscratchs);

		// P4 = S2 * T2
		double *P4 = S + lscratchs;
		lscratchs += hdimC;
		wino_mm(b, c, d, hm, hk, hn, S2, T2, P4, S, &lscratchs);

		// P5 = S3 * T3
		double *P5 = S + lscratchs;
		lscratchs += hdimC;
		wino_mm(b, c, d, hm, hk, hn, S3, T3, P5, S, &lscratchs);

		// P6 = S3 * T3
		double *P6 = C12;
		wino_mm(b, c, d, hm, hk, hn, S4, B22, P6, S, &lscratchs);

		// P7 = A22 * T4
		double *P7 = C21;
		wino_mm(b, c, d, hm, hk, hn, A22, T4, P7, S, &lscratchs);

		double *U2 = S + lscratchs;
		double *U3 = U2 + hdimC;
		lscratchs += hdimC << 1;

		for ( i = 0; i < hdimC; i += bC ) {
			// C11 = P1 + P2, U2 = P1 + P4, U3 = U2 + P5
			wino_madd4(bC, &P2[i], &P4[i], &P5[i], &C11[i], &U2[i], &U3[i]);
			// C21 = U3 + P7, C22 = U3 + P3, U6 = U2 + P3, C12 = U6 + P6 
			//wino_madd5(bC, &U3[i], &P3[i], &U2[i], &C21[i], &C22[i], &C12[i]);
			wino_madd5(bC, &U3[i], &U2[i], &C21[i], &C22[i], &C12[i]);
		}

		*scratchs = lscratchs;
	}
}

